/**
 * MCS Media Computer Software
 * Copyright 2018 by Wilfried Klaas
 * Project: SPSEmulator
 * File: Not.java
 * EMail: W.Klaas@gmx.de
 * Created: 25.11.2018 wklaa_000
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 */
package de.mcs.tools.sps.mnemonic;

import org.apache.commons.lang3.StringUtils;

import de.mcs.tools.sps.exceptions.IllegalArgument;
import de.mcs.tools.sps.exceptions.SyntaxError;

/**
 * CALL #x: Jump to address #x + 16 * PAGE and return on RTR
 * 
 * @author wklaa_000
 *
 */
public class CALL extends AbstractMnemonic implements Mnemonic {

  byte value;
  private boolean isLabel;

  public CALL(String line) throws SyntaxError {
    super(line);
    value = 0;
    isLabel = false;
  }

  @Override
  public void checkArgument() throws SyntaxError {
    if (StringUtils.isEmpty(getArgument())) {
      throw new SyntaxError(getLineNumber(),
          String.format("missing argument for %s.", this.getClass().getSimpleName()));
    }
    String argument = getArgument().trim();
    if (argument.startsWith(":")) {
      isLabel = true;
      return;
    }
    int myValue = getArgumentAsNumber();
    if ((myValue < 0) || (myValue > 15)) {
      throw new IllegalArgument(getLineNumber(),
          String.format("argument %s is not in range 0..15 for %s.", getArgument(), this.getClass().getSimpleName()));
    }
    value = (byte) (myValue & 0xFF);
  }

  @Override
  public int getByte() {
    return 0xD0 + value;
  }

  @Override
  public boolean isLabel() {
    return isLabel;
  }

  @Override
  public boolean isUsingPage() {
    return true;
  }

  @Override
  public boolean hasArgument() {
    return true;
  }
}
